/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.Parameter;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.core.SynapseEnvironment;
import org.wso2.esb.statistics.StatisticsReporterThread;
import org.wso2.esb.transport.jetty.JettyServer;
import org.wso2.utils.ServerConfiguration;
import org.wso2.utils.security.CryptoUtil;
import org.wso2.utils.transport.http.HttpServerManager;

import java.io.File;
/*
 *  ESB Startup Class
 */
public class ServiceBus {

    private static Log log = LogFactory.getLog(ServiceBus.class);

    //ESB lifecycle commads
    public static final String COMMAND_START  = "START";
    public static final String COMMAND_RUN    = "RUN";
    public static final String COMMAND_STOP   = "STOP";
    public static final String COMMAND_STATUS = "STATUS";
    public static final String COMMAND_HELP   = "HELP";

    /** The Jetty server used internally to host the admin console */
    private JettyServer jettyServer;

    private static ServiceBusConfiguration esbConfiguration;
    /** The thread which reads statistics from memory and write them to the DB */
    private StatisticsReporterThread statisticsReporterThread;
    private String axis2RepoLocation;
    private ServiceBusManager serviceBusManager;
    private static boolean isShutdownHookAdded;

    public ServiceBus() {
        if (System.getProperty("esb.home") == null) {
            System.setProperty("esb.home", ".");
        }
        esbConfiguration = ServiceBusConfiguration.getServiceBusInstance();
        serviceBusManager = new ServiceBusManager(esbConfiguration);
    }

    public void startServer() throws ServiceBusException {

        try {
            serviceBusManager.start();
        } catch (AxisFault axisFault) {
            log.fatal("WSO2 ESB failed to start", axisFault);
            throw new ServiceBusException(axisFault);
        }

        startListeners();
        /*try {
            HttpServerManager.startServer(org.wso2.esb.transport.jetty.Constants.HTTP_SERVER);
            HttpServerManager.startServer(org.wso2.esb.transport.jetty.Constants.HTTPS_SERVER);
        } catch (Exception e) {
            log.fatal("Error starting http/s servers", e);
            throw new ServiceBusException(e);
        }*/
    }

    public void stopServer() throws ServiceBusException {
        stopListeners();
    }

    private void startListeners() throws ServiceBusException {

        jettyServer = new JettyServer(serviceBusManager.configurationContext);
        try {
            jettyServer.start();
        } catch (AxisFault axisFault) {
            throw new ServiceBusException(axisFault);
        }

        Parameter synapseEnvParameter = serviceBusManager.configurationContext.
                getAxisConfiguration().getParameter(org.apache.synapse.Constants.SYNAPSE_ENV);
        if (synapseEnvParameter != null) {
            try {
                String statisticsPersistenceInterval = esbConfiguration.
                        getFirstProperty("StatisticsPersistenceInterval");
                SynapseEnvironment synEnv = (SynapseEnvironment) synapseEnvParameter.getValue();
                statisticsReporterThread = new StatisticsReporterThread(synEnv);
                statisticsReporterThread.setDelay(Long.parseLong(statisticsPersistenceInterval));
                statisticsReporterThread.start();
            } catch (Exception e) {
                throw new ServiceBusException(e);
            }
        }
    }

    //THIS METHOD ONLY FOR TESTING PURPOSES.
    public void startListener(final int port) throws AxisFault, ServiceBusException {
        serviceBusManager.start();
        jettyServer = new JettyServer(serviceBusManager.configurationContext);
        jettyServer.start();
    }

    private void stopListeners() throws ServiceBusException {

        if (statisticsReporterThread != null) {
            statisticsReporterThread.shutdown();
        }

        // To shutdown derby network server
        try {
            serviceBusManager.stopDataBaseServer();
        } catch (Exception e) {
            log.fatal(e);
        }

        if (jettyServer != null) {
            try {
                jettyServer.stop();
            } catch (AxisFault axisFault) {
                log.fatal("ServiceBus#stopListener encounted an error", axisFault);
                throw new ServiceBusException("ServiceBus#stopListener encounted an error");
            }
        } else {
            log.fatal("ServiceBus#stopListener encounted an error");
            throw new ServiceBusException(
                    "ServiceBus#stopListener encounted an error");
        }
    }

    public void status() {
        //serviceBusManager.status();
    }

    public void shutdown() throws ServiceBusException {
        stopListeners();
    }

    public static void main(String[] args) {//throws Exception {
        String cmd;
        if (args.length >= 1) {
            cmd = args[0];
        } else {
            log.error("ERROR :: Wrong number of arguments.");
            printUsages();
            return;
        }

        long before = System.currentTimeMillis();
        final ServiceBus esb = new ServiceBus();

        esb.setAxis2RepoLocation(esbConfiguration.getFirstProperty(
                ServerConfiguration.AXIS2_CONFIG_REPO_LOCATION));
        CryptoUtil cryptoUtil =
                new CryptoUtil(System.getProperty("esb.home") + File.separator +
                        esbConfiguration.getFirstProperty("Security.KeyStore.Location"),
                        esbConfiguration.getFirstProperty("Security.KeyStore.Password"),
                        esbConfiguration.getFirstProperty("Security.KeyStore.KeyAlias"),
                        esbConfiguration.getFirstProperty("Security.KeyStore.KeyPassword"),
                        esbConfiguration.getFirstProperty("Security.KeyStore.Type"));


        if (cmd.toUpperCase().equals(COMMAND_START)) {
            // when starting the server with start command, a server socket starts to
            // listen to the incomming commands. but no shut down hooks are registered
            log.info("Using Repository " + new File(".").getAbsolutePath() +
                    File.separator + esb.getAxis2RepoLocation());
            startServiceBusController(esb);
            try {
                esb.startServer();
            } catch (ServiceBusException e) {
                log.fatal(e);
            }
            log.info("WSO2 ESB started in " +
                    (System.currentTimeMillis() - before) + " ms");

        } else if (cmd.toUpperCase().equals(COMMAND_RUN)) {

            // when starting the server with start command, a server socket starts to
            // listen to the incomming commands. but no shut down hooks are registered
            log.info("Using Repository " + new File(".").getAbsolutePath() +
                    File.separator + esb.getAxis2RepoLocation());

            startServiceBusController(esb);
            try {
                esb.startServer();
                esb.addShutdownHook();
            } catch (ServiceBusException e) {
                log.fatal(e);
            }
            log.info("WSO2 ESB server started in " +
                    (System.currentTimeMillis() - before) + " ms");

        } else if (cmd.toUpperCase().equals(COMMAND_STOP)) {
            ServiceBusController esbController =
                    new ServiceBusController(esb);
            int serverPort = Integer.parseInt(esbConfiguration.
                    getFirstProperty(
                    ServerConfiguration.COMMAND_LISTENER_PORT));
            esbController.setPort(serverPort);
            esbController.shutDownServer();

        } else if (cmd.toUpperCase().equals(COMMAND_STATUS)) {
            esb.status();
        } else if (cmd.toUpperCase().equals(COMMAND_HELP)) {
            printUsages();
        }

    }

    private static void startServiceBusController(ServiceBus esb) {
        ServiceBusController esbController =
                new ServiceBusController(esb);
        int serverPort = Integer.parseInt(esbConfiguration.
                getFirstProperty(
                ServerConfiguration.COMMAND_LISTENER_PORT));
        esbController.setPort(serverPort);
        Thread t = new Thread(esbController);
        t.start();

    }


    private static void printUsages() {
        System.out.println(
                "======================================================");
        System.out.println(
                "=                WSO2 ESB Help                       =");
        System.out.println(
                "======================================================");
        System.out.println();
        System.out.println(" Usage : ESB <action>");
        System.out.println();
        System.out.println("<action> \t\t : Current actions supported - " +
                COMMAND_RUN + ", " +
                COMMAND_START + ", " +
                COMMAND_STATUS + "," +
                COMMAND_STOP + "," +
                COMMAND_HELP);
        System.out.println();
    }


    public void setAxis2RepoLocation(String axis2RepoLocation) {
        this.axis2RepoLocation = axis2RepoLocation;
    }

    private String getAxis2RepoLocation() {
        return this.axis2RepoLocation;
    }

    private void addShutdownHook() {
        if (isShutdownHookAdded) {
            return;
        }
        isShutdownHookAdded = true;
        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() {
                log.info("[ESB] Shutting down WSO2 ESB ...");
                try {
                    ServiceBus.this.shutdown();
                    log.info("[ESB] Shutting down completed");
                    log.info("[ESB] Halting JVM");
                } catch (Exception e) {
                    log.warn("[ESB] Error occurred while " +
                            "shutting down WSO2 ESB : " + e);
                }
            }
        });
    }

    public ConfigurationContext getConfigurationContext() {
        return this.serviceBusManager.configurationContext;
    }

}
