/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb.services;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.config.xml.ProxyServiceFactory;
import org.apache.synapse.config.xml.ProxyServiceSerializer;
import org.apache.synapse.core.axis2.ProxyService;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.esb.services.tos.ProxyData;

import javax.xml.namespace.QName;
import java.util.*;

/**
 * This is a POJO for the Proxy Service admin service
 */
public class ProxyServiceAdmin extends AbstractESBAdmin {

    private static Log log = LogFactory.getLog(ProxyServiceAdmin.class);

    /**
     * @return
     * @throws AxisFault
     */
    public ProxyData[] proxyData() throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Collection proxyCollection = synapseConfiguration.getProxyServices();
        ArrayList proxyServiceList = new ArrayList(1);
        for (Iterator ite = proxyCollection.iterator();
             ite.hasNext();) {
            Object obj = ite.next();
            if (obj instanceof ProxyService) {
                ProxyService proxy = (ProxyService) obj;
                ProxyData data = new ProxyData();
                data.setName(proxy.getName());
                if (proxy.getTargetInSequence() != null || proxy.getTargetOutSequence() != null) {
                    data.setTarget(proxy.getTargetInSequence());
                } else {
                    data.setTarget(proxy.getTargetEndpoint());
                }
                data.setRunning(proxy.isRunning());
                if(proxy.getStatisticsEnable() == org.apache.synapse.Constants.STATISTICS_ON) {
                    data.setEnableStatistics(true);
                } else {
                    data.setEnableStatistics(false);
                }
                if(proxy.getTraceState() == org.apache.synapse.Constants.TRACING_ON) {
                    data.setEnableTracing(true);
                } else {
                    data.setEnableTracing(false);
                }
                if(proxy.getWsdlURI() != null ||
                        proxy.getWSDLKey() != null || proxy.getInLineWSDL() != null) {
                    data.setWsdlAvailabel(true);
                } else {
                    data.setWsdlAvailabel(false);
                }
                if(proxy.getServiceLevelPolicies().isEmpty()) {
                    data.setPolicyAvailabel(false);
                } else {
                    data.setPolicyAvailabel(true);
                }
                proxyServiceList.add(data);
            }
        }

        Collections.sort(proxyServiceList, new Comparator() {
            public int compare(Object o1, Object o2) {
                return ((ProxyData) o1).getName().compareTo(
                    ((ProxyData)o2).getName());
            }
        });

        return (ProxyData[]) proxyServiceList
                .toArray(new ProxyData[proxyServiceList.size()]);
    }

    /**
     * 
     * @param proxyName
     * @throws AxisFault
     */
    public void enableStatistics(String proxyName) throws AxisFault {
        
        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setStatisticsEnable(org.apache.synapse.Constants.STATISTICS_ON);
            log.info("Enabled statistics on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to enable statistics for proxy service " + proxyName, syne);
        }
    }

    /**
     * 
     * @param proxyName
     * @throws AxisFault
     */
    public void disableStatistics(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setStatisticsEnable(org.apache.synapse.Constants.STATISTICS_OFF);
            log.info("Disabled statistics on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to disable statistics for proxy service " + proxyName, syne);
        }
    }

    /**
     * 
     * @param proxyName
     * @throws AxisFault
     */
    public void enableTracing(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setTraceState(org.apache.synapse.Constants.TRACING_ON);
            log.info("Enabled tracing on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to enable tracing for proxy service " + proxyName, syne);
        }
    }

    /**
     * 
     * @param proxyName
     * @throws AxisFault
     */
    public void disableTracing(String proxyName) throws AxisFault {

        try {
            getSynapseConfiguration().getProxyService(
                    proxyName).setTraceState(org.apache.synapse.Constants.TRACING_OFF);
            log.info("Disabled tracing on proxy service : " + proxyName);
        } catch (SynapseException syne) {
            handleFault(log, "Unable to disable tracing for proxy service " + proxyName, syne);
        }
    }

    /**
     * @param proxyName
     * @return
     * @throws AxisFault
     */
    public OMElement getProxyServiceElement(String proxyName) throws AxisFault {
        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        try {
            if (synapseConfiguration.getProxyService(proxyName) != null) {
                return ProxyServiceSerializer.serializeProxy(null,
                        synapseConfiguration.getProxyService(proxyName));
            } else {
                handleFault(log,"A proxy service named : " + proxyName + " does not exist", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to get the proxy service definition for : " + proxyName, syne);
        }
        return null;
    }

    /**
     * Add a proxy service described by the given OMElement
     *
     * @param proxyServiceElement
     * @throws AxisFault if the element is not an proxy service or if a proxy service wiht the
     *                   same name exists
     */
    public void addProxyService(OMElement proxyServiceElement) throws AxisFault {
        try {
            if (proxyServiceElement.getQName().getLocalPart()
                    .equals(org.apache.synapse.config.xml.Constants.PROXY_ELT.getLocalPart())) {
                String proxyName = proxyServiceElement.getAttributeValue(new QName("name"));
                if (getSynapseConfiguration().getProxyService(proxyName) != null) {
                    handleFault(log, "A proxy service named : " + proxyName + " already exists", null);

                } else {
                    ProxyService proxy = ProxyServiceFactory.createProxy(proxyServiceElement);

                    try {
                        proxy.buildAxisService(getSynapseConfiguration(),
                            getSynapseConfiguration().getAxisConfiguration());

                        log.info("Added proxy service : " + proxyName);
                        if(!proxy.isStartOnLoad()) {
                            proxy.stop(getSynapseConfiguration());
                        }
                        getSynapseConfiguration().addProxyService(proxy.getName(), proxy);

                    } catch (AxisFault af) {
                        log.error("Unable to add Proxy service : " + proxy.getName(), af);
                        if (getAxisConfig().getService(proxy.getName()) != null) {
                            try {
                                getAxisConfig().removeService(proxy.getName());
                            } catch (Exception ignore) {}
                        }
                        throw new AxisFault("Unable to add Proxy service : " + proxy.getName(), af);
                    }
                }
            } else {
                handleFault(log, "Unable to add proxy service. Invalid definition", null);
            }
        } catch (SynapseException syne) {
            handleFault(log, "Unable to add proxy service. Invalid definition", syne);
        }
    }

    /**
     * Saves the proxy service to the SynapseConfiguration
     *
     * @param proxyServiceElement
     * @throws AxisFault
     */
    public void saveProxyService(OMElement proxyServiceElement) throws AxisFault {
        
        try {
            if (proxyServiceElement.getQName().getLocalPart()
                    .equals(org.apache.synapse.config.xml.Constants.PROXY_ELT.getLocalPart())) {
                
                String proxyName = proxyServiceElement.getAttributeValue(new QName("name"));

                ProxyService currentProxy = getSynapseConfiguration().getProxyService(proxyName);
                boolean wasRunning = false;
                if (currentProxy == null) {
                    handleFault(log, "Unable to save proxy service : "
                        + proxyName + ". Does not exist", null);

                } else {
                    wasRunning = currentProxy.isRunning();
                    log.debug("Deleting existing proxy service : " + proxyName);
                    deleteProxyService(proxyName);

                    try {
                        log.debug("Adding proxy service : " + proxyName);
                        addProxyService(proxyServiceElement);
                        log.info("Added added proxy service : " + proxyName);

                        if (!wasRunning &&
                            getSynapseConfiguration().getProxyService(proxyName).isRunning()) {
                            getSynapseConfiguration().getProxyService(
                                    proxyName).stop(getSynapseConfiguration());
                        } else if (wasRunning &&
                            !getSynapseConfiguration().getProxyService(proxyName).isRunning()) {
                            getSynapseConfiguration().getProxyService(
                                    proxyName).start(getSynapseConfiguration());
                        }
                    } catch (Exception e) {

                        log.error("Unable to save changes made for the proxy service : "
                            + proxyName + ". Restoring the existing proxy..");
                        getSynapseConfiguration().addProxyService(proxyName, currentProxy);
                        currentProxy.buildAxisService(getSynapseConfiguration(),
                                getSynapseConfiguration().getAxisConfiguration());
                        if (!wasRunning) {
                            currentProxy.stop(getSynapseConfiguration());
                        } else {
                            currentProxy.start(getSynapseConfiguration());
                        }
                        
                        handleFault(log, "Unable to save changes made for the proxy service : "
                            + proxyName + ". Restored the existing proxy...", e);
                    }
                }
                
            } else {
                handleFault(log, "Unable to save proxy service. Invalid definition", null);
            }
            
        } catch (SynapseException syne) {
            handleFault(log, "Unable to save proxy service. Invalid definition", syne);
        }
    }

    /**
     * Deletes a proxy service from the synapse configuration
     *
     * @param proxyServiceName
     * @throws AxisFault if the proxy service name given is not existent in the
     *                   synapse configuration
     */
    public void deleteProxyService(String proxyServiceName) throws AxisFault {
        try {
            log.debug("Deleting proxy service : " + proxyServiceName);
            SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
            synapseConfiguration.removeProxyService(proxyServiceName);
            log.info("Proxy service : " + proxyServiceName + " deleted");
        } catch (SynapseException syne) {
            handleFault(log, "Unable to delete proxy service : " + proxyServiceName, syne);
        }
    }

    /**
     * Get the availabel transport names from the AxisConfiguration
     *
     * @return String array of availabel transport names
     * @throws AxisFault if there is an error
     */
    public String[] getAvailabelTransports() throws AxisFault {
        Object[] transports = getAxisConfig().getTransportsIn().keySet().toArray();
        String[] ret = new String[transports.length];
        for (int i = 0; i < transports.length; i++) {
            ret[i] = (String) transports[i];
        }
        return ret;
    }

    /**
     * Get the availabel sequences from the SynapseConfiguration
     *
     * @return String array of availabel sequence names
     * @throws AxisFault if there is an error
     */
    public String[] getAvailabelSequences() throws AxisFault {
        Object[] sequences = getSynapseConfiguration().getDefinedSequences().keySet().toArray();
        String[] ret = new String[sequences.length];
        for (int i = 0; i < sequences.length; i++) {
            ret[i] = (String) sequences[i];
        }
        return ret;
    }

    /**
     * Get the availabel endpoints from the SynapseConfiguration
     *
     * @return String array of availabel endpoint names
     * @throws AxisFault if there is an error
     */
    public String[] getAvailabelEndpoints() throws AxisFault {
        Object[] endpoints = getSynapseConfiguration().getDefinedEndpoints().keySet().toArray();
        String[] ret = new String[endpoints.length];
        for (int i = 0; i < endpoints.length; i++) {
            ret[i] = (String) endpoints[i];
        }
        return ret;
    }

    public void startProxyService(String proxyServiceName) throws AxisFault {
        log.debug("Starting/Re-starting proxy service : " + proxyServiceName);
        getSynapseConfiguration().getProxyService(proxyServiceName)
                .start(getSynapseConfiguration());
        log.info("Started/Re-started proxy service : " + proxyServiceName);
    }

    public void stopProxyService(String proxyServiceName) throws AxisFault {
        log.debug("Stopping proxy service : " + proxyServiceName);
        getSynapseConfiguration().getProxyService(proxyServiceName)
                .stop(getSynapseConfiguration());
        log.info("Stopped proxy service : " + proxyServiceName);
    }
}
