/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb.services;

import org.apache.axis2.AxisFault;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.synapse.SynapseException;
import org.apache.synapse.config.SynapseConfiguration;
import org.apache.synapse.registry.Registry;
import org.apache.synapse.registry.RegistryEntry;
import org.apache.synapse.registry.url.SimpleURLRegistry;
import org.apache.axiom.om.*;
import org.wso2.esb.registry.ESBRegistry;
import org.wso2.esb.registry.ESBRegistryEntry;
import org.wso2.esb.Constants;

import javax.xml.namespace.QName;
import java.util.ArrayList;
import java.io.File;

public class RegistryAdmin extends AbstractESBAdmin {

    private static final Log log = LogFactory.getLog(RegistryAdmin.class);

    public String getDefaultRegistryName() throws AxisFault {
        Registry reg = getDefaultRegistry();
        if (reg == null) {
            return "Registry is not defined";
        }
        
        String regClass = reg.getProviderClass();

        if (ESBRegistry.class.getName().equals(regClass)) {
            return "ESB Registry";
        } else if (SimpleURLRegistry.class.getName().equals(regClass)) {
            return "Default Synapse Registry";
        } else {
            return "Unknown Registry";
        }
    }

    public String[] getDescendants() throws AxisFault {

        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Registry reg = synapseConfiguration.getRegistry();

        ArrayList keyList = new ArrayList();

        try {
            RegistryEntry[] entries = reg.getDescendants(null);
            for(int i=0; i<entries.length; i++) {
                keyList.add(entries[i].getKey());
            }
        } catch (SynapseException e) {
            handleFault(log, "Couldn't get descendants from registry", e);
        }

        String[] keys = new String[keyList.size()];
        for(int i=0; i<keyList.size(); i++) {
            keys[i] = (String)keyList.get(i);
        }

        return keys;
    }

    /**
     * Returns the children of the given registry entry.
     *
     * @param key
     *
     * @return
     * <children>
     * <regEntry><key>key</key><type>file/folder</type></regEntry>
     * ...
     * </children>
     *
     * @throws AxisFault
     */
    public OMElement getChildren(String key) throws AxisFault {

        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMElement children = fac.createOMElement("children", null);

        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        Registry reg = synapseConfiguration.getRegistry();

        try {
            ESBRegistryEntry entry = new ESBRegistryEntry();
            // key = null ==> list the root level childrens
            if (key == null) {
                key = "";
            }
            entry.setKey(key);
            RegistryEntry[] entries = reg.getChildren(entry);
            if (entries != null) {
                for(int i=0; i<entries.length; i++) {
                    OMElement entryElement = fac.createOMElement("regEntry", null);

                    OMElement keyElement = fac.createOMElement("key", null);
                    keyElement.setText(entries[i].getKey());
                    entryElement.addChild(keyElement);

                    OMElement typeElement = fac.createOMElement("type", null);
                    if (entries[i].getType() != null && entries[i].getType() == Constants.folder) {
                        typeElement.setText("folder");
                    } else {
                        typeElement.setText("file");
                    }
                    entryElement.addChild(typeElement);

                    children.addChild(entryElement);
                }
            } else {
                handleFault(log, "Selected entry is a leaf level entry. " +
                    "No child entries are available for leaf entries", null);
            }
        } catch (SynapseException e) {
            handleFault(log, "Couldn't get child elements from registry", e);
        }
        return children;
    }

    /**
     * Gives the registry entry meta data (not the actual registry entry)
     *
     * @param key
     * @return
     * @throws AxisFault
     */
    public OMElement getRegistryEntry(String key) throws AxisFault {

        Registry reg = getDefaultRegistry();

        if (reg != null) {
            RegistryEntry entry = reg.getRegistryEntry(key);

            // initialize all values to defaults.
            String sExpiryTime = "0";

            if (entry != null) {
                // registry entry found. update values from that.
                // we are exposing cachable duration to outside in seconds and it is kept internally in milliseconds
                Long expiryTime = new Long(entry.getCachableDuration() / 1000);
                sExpiryTime = expiryTime.toString();
            }

            OMFactory fac = OMAbstractFactory.getOMFactory();
            OMElement regEntry = fac.createOMElement("regEntry", null);

            OMElement keyElement = fac.createOMElement("key", null);
            keyElement.setText(key);
            regEntry.addChild(keyElement);

            OMElement expElement = fac.createOMElement("expiryTime", null);
            expElement.setText(sExpiryTime);
            regEntry.addChild(expElement);

            return regEntry;

        } else {
            handleFault(log, "No registry found", null);
            return null;
        }
    }

    /**
     * Updates the registry entry data. Actual registry entry has to created either by uploading a
     * file or creating a new folder before calling this method.
     *
     * @param entryElement
     * @throws AxisFault
     */
    public void updateRegistryEntry(OMElement entryElement) throws AxisFault {

        Registry reg = getDefaultRegistry();

        if (reg != null) {

            ESBRegistryEntry entry = new ESBRegistryEntry();

            OMElement keyElement = entryElement.getFirstChildWithName(new QName("key"));
            if (keyElement != null) {
                entry.setKey(keyElement.getText());
            } else {
                handleFault(log, "All registry entries should have a key", null);
            }

            log.debug("Updating registry entry with key : " + keyElement.getText());
            OMElement expiryTime = entryElement.getFirstChildWithName(new QName("expiryTime"));
            if (expiryTime != null) {
                long time = 0;
                try {
                    time = new Long(expiryTime.getText()).longValue();
                } catch (NumberFormatException e) {
                    handleFault(log, "Cachable duration should be a valid number", e);
                }
                // we are exposing cachable duration to outside in seconds and it is kept internally in milliseconds
                entry.setCachableDuration(time * 1000);
            }

            ESBRegistry esbRegistry = (ESBRegistry) reg;
            esbRegistry.updateRegistryEntry(entry);
            log.info("Updated registry entry with key : " + keyElement.getText());

        } else {
            handleFault(log, "No registry found", null);
        }
    }

    /**
     * Creates new folder in the given parent folder.
     *
     * @param entryElement
     * <folder><folderName>new folder name</folderName><parent>parent folder name</parent></folder>
     *
     * @return OMElement
     * <folder>
     * <folderPath>full new folder name<folderPath><status>successfull/failed</status>
     * </folder>
     *
     * @throws AxisFault
     */
    public OMElement createFolder(OMElement entryElement) throws AxisFault {

        Registry reg = getDefaultRegistry();

        if (reg != null) {

            String newFolderName  = null;
            String parentName = null;

            OMElement entryNameElement
                    = entryElement.getFirstChildWithName(new QName("folderName"));
            OMElement parentNameElement
                    = entryElement.getFirstChildWithName(new QName("parent"));

            if (entryNameElement != null && parentNameElement != null) {
                newFolderName = entryNameElement.getText();
                parentName = parentNameElement.getText();
            } else {
                handleFault(log, "Invalid input data. Null entry name or parent", null);
            }

            log.debug("Creating folder : " + newFolderName + " under : " + parentName);
            ESBRegistry esbRegistry = (ESBRegistry) reg;

            try {
                esbRegistry.addResource(parentName, newFolderName, false);
                //esbRegistry.createFolder(parentName, newFolderName);

                OMFactory fac = OMAbstractFactory.getOMFactory();
                OMElement rFolder = fac.createOMElement("folder", null);

                OMElement path = fac.createOMElement("folderPath", null);
                path.setText(parentName + ESBRegistry.URL_SEPARATOR + newFolderName);
                rFolder.addChild(path);

                OMElement status = fac.createOMElement("status", null);
                status.setText("successfull");
                rFolder.addChild(status);

                log.info("Created folder : " + newFolderName + " under : " + parentName);
                return rFolder;

            } catch (Exception e) {
                handleFault(log, "Couldn't add new Registry entry: " + newFolderName, e);
            }

        } else {
            handleFault(log, "No registry found", null);
        }
        return null;
    }

    /**
     * Adds a new file to the registry.
     *
     * @param fileElement
     * <file>
     * <fileName>file name</fileName><parent>parent key</parent><content>xml content</content>
     * </file>
     * if <content> element is absent, creates an empty file.
     *
     * @return OMElement
     * <file><filePath>full file path</filePath><status>ok/failed</status></file>
     */
    public OMElement createFile(OMElement fileElement) throws AxisFault {
        Registry reg = getDefaultRegistry();

        if (reg != null) {

            String newFileName = null;
            String parentName = null;
            String content = null;

            OMElement entryNameElement = null;
            OMElement parentNameElement = null;
            OMElement contentWrapper = null;
            try {
                entryNameElement = fileElement.getFirstChildWithName(new QName("fileName"));
                parentNameElement = fileElement.getFirstChildWithName(new QName("parent"));
                contentWrapper = fileElement.getFirstChildWithName(new QName("content"));
            } catch (Exception e) {
                handleFault(log, "Invalid data. Content should be XML", e);
            }

            log.debug("Creating file : " + newFileName + " under : " + parentName);

            OMElement contentElement = null;
            if (contentWrapper != null) {
                contentElement = contentWrapper.getFirstElement();
                if (contentElement == null) {
                    //if the user has specified a <content> tag, he should provide a valid xml in it
                    handleFault(log, "Invalid data. Content should be XML", null);
                }
            }

            if (entryNameElement != null && parentNameElement != null && contentElement != null) {
                newFileName = entryNameElement.getText();
                parentName = parentNameElement.getText();
            } else {
                handleFault(log, "Unable to create resource. Resource name, content and parent" +
                    " directory must be specified", null);
            }

            ESBRegistry esbRegistry = (ESBRegistry) reg;

            try {
                esbRegistry.addResource(parentName, newFileName, true);
                esbRegistry.updateResource
                        (parentName + ESBRegistry.URL_SEPARATOR + newFileName, contentElement);
                //esbRegistry.createFile(parentName, newFileName, content);

                OMFactory fac = OMAbstractFactory.getOMFactory();
                OMElement rFile = fac.createOMElement("file", null);

                OMElement filePath = fac.createOMElement("filePath", null);
                filePath.setText(parentName + ESBRegistry.URL_SEPARATOR + newFileName);
                rFile.addChild(filePath);

                OMElement status = fac.createOMElement("status", null);
                status.setText("ok");
                rFile.addChild(status);

                log.info("Created file : " + newFileName + " under : " + parentName);
                return rFile;
                
            } catch (Exception e) {
                handleFault(log,
                    "Couldn't create new file: " + newFileName + " under : " + parentName, e);
            }

        } else {
            handleFault(log, "No registry found", null);
        }
        return null;
    }

    public void deleteRegistryResource(String key) throws AxisFault {

        log.debug("Removing entry with key : " + key);
        ESBRegistry esbRegistry = (ESBRegistry) getDefaultRegistry();
        esbRegistry.removeResource(key);
        log.info("Entry with key : " + key + " removed from the registry");
    }

    public OMElement getRegistryEntryValue(String key) throws AxisFault {

        // check if the request resource is a file or a folder
        // if it is a folder return a error message
        // if not look up the resource and append it to the value part

        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMElement entryElement = fac.createOMElement("regEntry", null);

        Registry reg = getDefaultRegistry();

        try {
            OMNode resourceNode = reg.lookup(key);

            OMElement keyElement = fac.createOMElement("key", null);
            keyElement.setText(key);
            entryElement.addChild(keyElement);

            OMElement valueElement = fac.createOMElement("value", null);
            valueElement.addChild(resourceNode);
            entryElement.addChild(valueElement);
        } catch(Exception e) {
            // add an error message to the body
            OMElement errElement = fac.createOMElement("error", null);
            errElement.setText("Couldn't read from the registry entry.");
            entryElement.addChild(errElement);
        }

        return entryElement;
    }

    /**
     * Updates the registry entry file.
     *
     * @param regEntryElement
     * <regEntry><key>registry key</key><value>xml contents of the file</value></regEntry>
     */
    public void updateRegistryEntryValue(OMElement regEntryElement) throws AxisFault {

        OMElement keyElement = null;
        OMElement valueWrapper = null;
        try {
            keyElement = regEntryElement.getFirstChildWithName(new QName("key"));
            valueWrapper = regEntryElement.getFirstChildWithName(new QName("value"));
        } catch (OMException e) {
            handleFault(log, "Error updating entry. Invalid data", e);
        }

        if (valueWrapper != null) {
            OMElement valueElement = null;
            try {
                valueElement = valueWrapper.getFirstElement();
            } catch (Exception e) {
                handleFault(log, "Error updating entry. Content should be XML", e);
            }
            if (keyElement != null && valueElement != null) {
                String key = keyElement.getText();

                log.debug("Updating registry entry for key : " + key);
                ESBRegistry reg = (ESBRegistry) getDefaultRegistry();

                try {
                    reg.updateResource(key, valueElement);
                    //reg.updateRegistryEntryValue(key, valueElement);
                    log.debug("Updated registry entry for key : " + key);
                } catch (Exception e) {
                    handleFault(log, "Error updating entry. Couldn't write to file: " + key, e);
                }
            }
        }

    }

    private Registry getDefaultRegistry() throws AxisFault {

        SynapseConfiguration synapseConfiguration = getSynapseConfiguration();
        return synapseConfiguration.getRegistry();
    }
}
