/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.esb.services.tos;

import org.apache.axis2.AxisFault;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.AxisServiceGroup;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.wso2.statistics.StatisticsConstants;
import org.wso2.statistics.Counter;
import org.wso2.statistics.module.StatisticsModule;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Iterator;

/*
 * Holds information about the server status which is displayed on the home page
 */
public class ServerStatus {

    private int responseCount;
    private int faultCount;
    private String serverName;
    private String serverStartTime;
    private String systemUpTime;
    private int requestCount = 0;
    private int services;
    private String freeMemory;
    private String totalMemory;
    private double avgResponseTime;
    private long minResponseTime;
    private long maxResponseTime;

    private SimpleDateFormat dateFormatter = new SimpleDateFormat(
            "yyyy-MM-dd HH:mm:ss");

    /**
     * Generate server status data reading necessary information from the
     * Axis2 configuration passed in
     *
     * @param axisConfig
     * @throws AxisFault
     */
    public ServerStatus(AxisConfiguration axisConfig) throws AxisFault {

        Parameter systemStartTime = axisConfig.getParameter(StatisticsConstants.SERVER_START_TIME);
        long startTime = ((Long) systemStartTime.getValue()).longValue();
        Date stTime = new Date(startTime);
        serverStartTime = dateFormatter.format(stTime);
        systemUpTime = (getTime((System.currentTimeMillis() - startTime) / 1000));

        Parameter globalCounter = axisConfig.getParameter(StatisticsConstants.GLOBAL_REQUEST_COUNTER);

        // need to handle responses as well, once the metrics are in order with Synapse.
        if (globalCounter != null) {
            Object value = globalCounter.getValue();
            if (value instanceof Counter) {
                requestCount = ((Counter) value).getCount();
            }
        }

        responseCount =
                ((org.wso2.statistics.Counter) axisConfig.
                        getParameter(StatisticsConstants.GLOBAL_RESPONSE_COUNTER).getValue()).
                        getCount();
        faultCount =
                ((org.wso2.statistics.Counter) axisConfig.
                        getParameter(StatisticsConstants.GLOBAL_FAULT_COUNTER).getValue()).
                        getCount();

        freeMemory  = formatMemoryValue(Runtime.getRuntime().freeMemory());
        totalMemory = formatMemoryValue(Runtime.getRuntime().totalMemory());

        AxisServiceGroup asg =
                axisConfig.getServiceGroup(org.wso2.esb.Constants.ADMIN_SERVICE_GROUP);

        int adminservicecout = 0;
        int activeServices = 0;
        int totalRequestForAdmin = 0;
        int totalResponseFromAdmin = 0;
        int totalFaultCountForAdmin = 0;

        Iterator iter = asg.getServices();
        while (iter.hasNext()) {

            AxisService axisService = (AxisService) iter.next();
            Parameter parameter =
                    axisService.getParameter(StatisticsConstants.SERVICE_REQUEST_COUNTER);

            if (parameter != null) {
                totalRequestForAdmin += ((Counter) parameter.getValue()).getCount();
            }

            Parameter faultCounterParameter =
                    axisService.getParameter(StatisticsConstants.SERVICE_FAULT_COUNTER);

            if (faultCounterParameter != null) {
                totalFaultCountForAdmin += ((Counter) faultCounterParameter.getValue()).getCount();
            }

            totalResponseFromAdmin = totalResponseFromAdmin + getServiceResponseCount(axisService);
            adminservicecout++;
        }

        Iterator services = axisConfig.getServices().values().iterator();

        while (services.hasNext()) {
            AxisService axisService = (AxisService) services.next();

            if (axisService.isClientSide()) {
                continue;
            }

            if (axisService.isActive()) {
                activeServices++;
            }
        }

        // counter includes Synapse Service as well
        // Following code only Shows the awailable proxy services. 
        requestCount  -= totalRequestForAdmin;
        responseCount -= totalResponseFromAdmin;
        faultCount    -= totalFaultCountForAdmin;
        this.services  = activeServices - adminservicecout - 1;
        serverName     = org.wso2.esb.Constants.ESB_INSTANCE;

        avgResponseTime = StatisticsModule.responseTimeProcessor.getAvgResponseTime();
        maxResponseTime = StatisticsModule.responseTimeProcessor.getMaxResponseTime();
        minResponseTime = StatisticsModule.responseTimeProcessor.getMinResponseTime();
    }

    private String formatMemoryValue(long value) {
        String formatterMemoryValue = "";

        if (value > 0) {
            if (((value / (1024 * 1024 * 1024)) >= 1)) {
                formatterMemoryValue = (value / (1024 * 1024 * 1024)) + " GB";
            } else if (((value / (1024 * 1024)) >= 1)) {
                formatterMemoryValue = (value / (1024 * 1024)) + " MB";
            } else if (((value / (1024)) >= 1)) {
                formatterMemoryValue = (value / (1024)) + " KB";
            } else {
                formatterMemoryValue = value + " bytes";
            }
        }

        return formatterMemoryValue;
    }

    private String getTime(long timeInSeconds) {
        long hours;
        long minutes;
        long seconds;
        hours = timeInSeconds / 3600;
        timeInSeconds = timeInSeconds - (hours * 3600);
        minutes = timeInSeconds / 60;
        timeInSeconds = timeInSeconds - (minutes * 60);
        seconds = timeInSeconds;

        return hours + " hour(s) " + minutes + " minute(s) " + seconds +
            " second(s)";
    }

    public int getServiceResponseCount(AxisService axisService) throws AxisFault {

        int count = 0;
        for (Iterator opIter = axisService.getOperations(); opIter.hasNext();) {
            AxisOperation axisOp = (AxisOperation) opIter.next();
            Parameter parameter = axisOp.getParameter(StatisticsConstants.OUT_OPERATION_COUNTER);
            if (parameter != null) {
                count += ((Counter) parameter.getValue()).getCount();
            }
        }
        return count;
    }

    public String getSystemUpTime() {
        return systemUpTime;
    }

    public String getServerStartTime() {
        return serverStartTime;
    }

    public String getServerName() {
        return serverName;
    }

    public int getRequestCount() {
        return requestCount;
    }

    public int getServices() {
        return services;
    }

    public String getFreeMemory() {
        return freeMemory;
    }

    public String getTotalMemory() {
        return totalMemory;
    }

    public double getAvgResponseTime() {
        return avgResponseTime;
    }

    public long getMinResponseTime() {
        return minResponseTime;
    }

    public long getMaxResponseTime() {
        return maxResponseTime;
    }

    public int getFaultCount() {
        return faultCount;
    }

    public void setFaultCount(int faultCount) {
        this.faultCount = faultCount;
    }

    public int getResponseCount() {
        return responseCount;
    }

    public void setResponseCount(int responseCount) {
        this.responseCount = responseCount;
    }
}
